
#ifndef Interpolator_h
#define Interpolator_h

#include "SolverTypes.h"
#include "Proof.h"

extern "C" {
#include "Solver_C.h"
}

#include <set>
#include <unordered_set>
#include <unordered_map>
#include <map>

using std::set;
using std::unordered_set;
using std::unordered_map;
using std::map;
using std::pair;

//=================================================================================================

class Interpolator : public ProofTraverser {

	//unordered_set<Var> variablesOfA;
	//unordered_set<Var> variablesOfB;
    unordered_set<Var> variablesOfOnlyA;
    unordered_set<Var> variablesOfOnlyB;
    unordered_set<Var> variablesOfBothAB;

    bool newRootsAreOfA;

    unordered_map<Var,int> variableOfAi;

    vec<vec<Lit> >            clauses;
    map<int,struct ItpNode*>  itpNodes;
    set<struct ItpNode *> anonymousNodes;

    struct ItpNode * lastItpNode;

    bool verboseProofLogging;
    int k;

public:

    Interpolator() : newRootsAreOfA (true)
                   , lastItpNode ((struct ItpNode *)NULL)
                   , verboseProofLogging (false)
				   , k (0)
                   {
                   }

    ~Interpolator() {

      /* variable sets */

      /*
      variablesOfA.~set<Var>();
      variablesOfB.~set<Var>();
      variablesOfOnlyA.~set<Var>();
      variablesOfOnlyB.~set<Var>();
      variablesOfBothAB.~set<Var>();
      */

      /* proof nodes */

      set<struct ItpNode *>::iterator sit;
      for (sit = anonymousNodes.begin(); sit != anonymousNodes.end(); sit++) {
        delete *sit;
      }

      map<int,struct ItpNode *>::iterator mit;
      for (mit = itpNodes.begin(); mit != itpNodes.end(); mit++) {
      	struct ItpNode *n = (struct ItpNode *) (*mit).second;
      	free(n->clause);
        //delete (*mit).second;
      	delete n;
      }

    }

    /* virtual methods from ProofTraverser class */
    void root                   (const vec<Lit>& c);
    void chain                  (const vec<ClauseId>& cs, const vec<Var>& xs);
    void deleted                (ClauseId c);

    void addVariableOfA         (Var v)    { variablesOfOnlyA.insert(v); }
    void addVariableOfB         (Var v)    { variablesOfOnlyB.insert(v); }

    void computeVariableLists   (void);
    void markProblemSplit       (int i)    {
    	if (i == -2) {
    		newRootsAreOfA = false;
    	} else {
    		k = i;
    	}
    }

    struct ItpNode * getLastItpNode (void) { return lastItpNode; }

    void printVariableLists     (void);
    void enableVerboseProofLogging (void)  { verboseProofLogging = true; }
    void disableVerboseProofLogging (void) { verboseProofLogging = false; }
    int isVariableOfB                       (int literal);

private:

    bool varOnlyBelongsToA      (Var v)    { return variablesOfOnlyA.end() != variablesOfOnlyA.find(v); }
    bool varOnlyBelongsToB      (Var v)    { return variablesOfOnlyB.end() != variablesOfOnlyB.find(v); }
    bool varBelongsToBothAandB  (Var v)    { return variablesOfBothAB.end() != variablesOfBothAB.find(v); }

    void deleteItpNode          (struct ItpNode * n) {
      if ((struct ItpNode *)NULL != n->parentNegative) deleteItpNode(n->parentNegative);
      if ((struct ItpNode *)NULL != n->parentPositive) deleteItpNode(n->parentPositive);
      delete n;
    }

};

//=================================================================================================

#endif

