
#ifndef __ISC_H__
#define __ISC_H__

#include "utils/utils.h"

#include "bmc/bmc.h"
#include "fsm/be/BeFsm.h"
#include "prop/Prop.h"
#include "trace/Trace.h"


/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/
typedef enum
{
  ISC_ALG_NOITP,            /* no interpolation (bmc style) */
  ISC_ALG_NOITPSPLIT,       /* no interpolation, module/environment split */
  ISC_ALG_STANDARD,         /* mcmillan */
  ISC_ALG_COMPOSITIONAL     /* nguyen */
}
Isc_algorithm;

typedef enum
{
  ISC_SPLITMODE_FVARIABLES,
  ISC_SPLITMODE_FMODULES,
  ISC_SPLITMODE_FCOMPONENT,
  ISC_SPLITMODE_NONE
}
Isc_splitmode;

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/
EXTERN void Isc_init ARGS((void));
EXTERN void Isc_init_cmd ARGS((void));
EXTERN void Isc_reset ARGS((void));
EXTERN void Isc_quit ARGS((void));

/* wrapper function for verification algorithms */

EXTERN int Isc_CheckInvar ARGS((Prop_ptr invarprop, Isc_algorithm alg,
      char* module_name, Isc_splitmode split_mode,
      boolean accumulate_env_trans, boolean xor_prop_violation));

/* verification algorithms */

EXTERN Bmc_result Isc_check_invar_noitp_algorithm ARGS((BeFsm_ptr be_fsm,
                                       node_ptr binvarspec,
                                       int max_k,
                                       Trace_ptr* trace,
                                       boolean xor_prop_violation,
                                       NodeList_ptr trace_symbols));

EXTERN Bmc_result Isc_check_invar_noitpsplit_algorithm ARGS((BeFsm_ptr be_fsm_monolithic,
                                                char* module_name,
                                                Isc_splitmode split_mode,
                                                node_ptr binvarspec,
                                                int max_k,
                                                Trace_ptr* trace,
                                                boolean xor_prop_violation,
                                                NodeList_ptr trace_symbols));

EXTERN Bmc_result Isc_check_invar_mcmillan_algorithm ARGS((BeFsm_ptr be_fsm,
                                              node_ptr binvarspec,
                                              Trace_ptr* trace,
                                              boolean xor_prop_violation,
                                              NodeList_ptr trace_symbols));

Bmc_result Isc_check_invar_nguyen_algorithm ARGS((BeFsm_ptr be_fsm_monolithic,
                                            char* module_name,
                                            Isc_splitmode split_mode,
                                            node_ptr binvarspec,
                                            Trace_ptr* trace,
                                            boolean accumulate_env_trans,
                                            boolean xor_prop_violation,
                                            NodeList_ptr trace_symbols));

#endif /* __ISC_H__ */

