
#include "isc.h"
#include "iscInt.h"

#include "cmd/cmd.h"
#include "prop/PropDb.h"
#include "prop/propPkg.h"

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
#define HAVE_ARG_PROPERTY \
  (strlen(str_property_idx) > 0  || strlen(str_formula) > 0)

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int UsageIscCheckInvar ARGS((void));

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Initializes the commands provided by this package]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void Isc_init_cmd(void)
{
  Cmd_CommandAdd("isc_check_invar", CommandIscCheckInvar, 0, true);
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           []

  CommandName        [isc_check_invar]            

  CommandSynopsis    []  

  CommandArguments   []  

  CommandDescription []  

  SideEffects        []

******************************************************************************/
int CommandIscCheckInvar(int argc, char **argv)
{
  /* parameters of Isc_CheckInvar() */

  Prop_ptr invarprop = PROP(NULL);
  Isc_algorithm alg = ISC_ALG_NOITP;
  Isc_splitmode split_mode = ISC_SPLITMODE_NONE;

  /* helpers for parsing command line arguments */

  int prop_no = -1;
  int c = 0;
  boolean have_arg_alg = false;

  char str_property_idx[MAX_STR+1];
  char str_formula[MAX_STR+1];
  char str_module_name[MAX_STR+1];
  char str_split_mode[MAX_STR+1];

  boolean have_accumulate_env_trans = false;
  boolean have_xor_prop_violation = false;

  str_property_idx[0] = '\0';
  str_formula[0] = '\0';
  str_module_name[0] = '\0';
  str_split_mode[0] = '\0';

  /* parse command line arguments */

  util_getopt_reset();
  while ((c = util_getopt(argc,argv,"a:n:p:m:s:cbxh")) != EOF) {
    switch (c) {
    case 'a':
      {
        if (have_arg_alg) return UsageIscCheckInvar();

        if (0 == strcmp(util_optarg, "standard")) {
          alg = ISC_ALG_STANDARD;
        }
        else if (0 == strcmp(util_optarg, "compositional")) {
          alg = ISC_ALG_COMPOSITIONAL;
        }

        #if ISC_DEBUG
        else if (0 == strcmp(util_optarg, "noitp")) {
          alg = ISC_ALG_NOITP;
        }
        else if (0 == strcmp(util_optarg, "noitpsplit")) {
          alg = ISC_ALG_NOITPSPLIT;
        }
        #endif

        else {
          return UsageIscCheckInvar();
        }

        have_arg_alg = true;
        break;
      }
    case 'n':
      {
        if (HAVE_ARG_PROPERTY) return UsageIscCheckInvar();

        if (strlen(util_optarg) > MAX_STR || strlen(util_optarg) < 1) {
          fprintf(nusmv_stderr, "Invalid string length for argument '-n'.\n");
          return 1;
        }

        strcpy(str_property_idx, util_optarg);
        break;
      }
    case 'p':
      {
        if (HAVE_ARG_PROPERTY) return UsageIscCheckInvar();

        if (strlen(util_optarg) > MAX_STR || strlen(util_optarg) < 1) {
          fprintf(nusmv_stderr, "Invalid string length for argument '-p'.\n");
          return 1;
        }

        strcpy(str_formula, util_optarg);
        break;
      }
    case 'm':
      {
        if (strlen(str_module_name) > 0) return UsageIscCheckInvar();
        if (ISC_SPLITMODE_NONE != split_mode) return UsageIscCheckInvar();

        if (strlen(util_optarg) > MAX_STR || strlen(util_optarg) < 1) {
          fprintf(nusmv_stderr, "Invalid string length for argument '-p'.\n");
          return 1;
        }

        strcpy(str_module_name, util_optarg);
        break;
      }
    case 's':
      {
        if (strlen(str_module_name) > 0) return UsageIscCheckInvar();
        if (ISC_SPLITMODE_NONE != split_mode) return UsageIscCheckInvar();

        if (0 == strcmp(util_optarg, "formula_variables")) {
          split_mode = ISC_SPLITMODE_FVARIABLES;
        }
        else if (0 == strcmp(util_optarg, "formula_modules")) {
          split_mode = ISC_SPLITMODE_FMODULES;
				}
        else if (0 == strcmp(util_optarg, "formula_component")) {
          split_mode = ISC_SPLITMODE_FCOMPONENT;
				}
        else {
          return UsageIscCheckInvar();
        }

        break;
      }
    case 'c':
      {
        have_accumulate_env_trans = true;
        break;
      }
    case 'x':
      {
        have_xor_prop_violation = true;
        break;
      }
    case 'h':
    default:
      return UsageIscCheckInvar();
    }
  }

  if (argc != util_optind) return UsageIscCheckInvar();

  /* preconditions */

  if ((ISC_SPLITMODE_NONE != split_mode || strlen(str_module_name) > 0) &&
      (alg == ISC_ALG_NOITP || alg == ISC_ALG_STANDARD)) {

    fprintf(nusmv_stderr, "A module name or split mode can only be given for "
        #if ISC_DEBUG
        "the algorithms 'noitpsplit' and 'compositional'.\n"
        #else
        "the algorithm 'compositional'.\n"
        #endif
        );
    return 1;
  }

  if (ISC_SPLITMODE_NONE == split_mode && strlen(str_module_name) == 0 &&
      alg != ISC_ALG_NOITP && alg != ISC_ALG_STANDARD) {
    fprintf(nusmv_stderr, "A module name (option -m) or a split mode (option -s) "
        #if ISC_DEBUG
        "must be given for the algorithms 'noitpsplit' and 'compositional'.\n"
        #else
        "must be given for the algorithm 'compositional'.\n"
        #endif
        );
    return 1;
  }

  #warning "TODO: check if there are at least two modules if splitting is used"

  if (!have_arg_alg) {
    fprintf(nusmv_stderr, "No algorithm was selected.\n");
    return 1;
  }

  if (!HAVE_ARG_PROPERTY) {
    fprintf(nusmv_stderr, "No property has been selected.\n");
    return 1;
  }

  if (have_accumulate_env_trans && alg != ISC_ALG_COMPOSITIONAL) {
    return UsageIscCheckInvar();
  }

  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf(nusmv_stderr, "The bmc environment must be set up before. Use the "
        "\"bmc_setup\" command.\n");
    return 1;
  }

  /* retrieve property */
  
  if (strlen(str_property_idx) > 0) {
    prop_no = PropDb_get_prop_index_from_string(PropPkg_get_prop_database(),
        str_property_idx);

    if (prop_no == -1) {
      fprintf(nusmv_stderr, "Property at index \"%s\"cannot be retrieved.\n",
          str_property_idx);
      return 1;
    }
  }

  else if (strlen(str_formula) > 0) {
    prop_no = PropDb_prop_parse_and_add(PropPkg_get_prop_database(),
				    Compile_get_global_symb_table(), str_formula, Prop_Invar);

    if (prop_no == -1) {
      fprintf(nusmv_stderr, "Property \"%s\" is invalid.\n", str_formula);
      return 1;
    }
  }

  else error_unreachable_code();

  invarprop = PropDb_get_prop_at_index(PropPkg_get_prop_database(), prop_no);
  nusmv_assert(PROP(NULL) != invarprop);

  /* check preconditions */

  if (Prop_Invar != Prop_get_type(invarprop)) {
    fprintf(nusmv_stderr, "The selected property is not an INVAR property.\n");
    return 1;
  }

  if (Prop_Unchecked != Prop_get_status(invarprop)) {
    fprintf(nusmv_stdout, "This property has already been verified.\n");
    return 0;
  }

  /* call wrapper function */

  return Isc_CheckInvar(invarprop, alg, str_module_name, split_mode,
      have_accumulate_env_trans, have_xor_prop_violation);
}

static int UsageIscCheckInvar()
{
  fprintf(nusmv_stderr,
      "\nusage: isc_check_invar -a \"algorithm\" (-n idx | -p \"formula\") [-m \"module\"]\n"
      "                       [-s \"splitmode\"] [-c] [-x] [-h]\n\n"
      "  -a \"algorithm\"\tUses the specified algorithm.\n"
      #if ISC_DEBUG
      "                  \tValid values are: noitp, noitpsplit, standard, compositional.\n"
      #else
      "                  \tValid values are: standard, compositional.\n"
      #endif
      "  -n idx          \tChecks the INVAR property specified with <idx>.\n"
      "  -p \"formula\"  \tChecks the specified INVAR property.\n"
      "  -m \"module\"   \tModule name for module/environment split.\n"
      "  -s \"splitmode\"\tValid values are: formula_modules, formula_variables,\n"
      "                  \t                  formula_component.\n"
      "  -c              \tAccumulate approximate environment transition\n"
      "                  \t(only with '-a compositional').\n"
      "  -x              \tChecks wether INVAR is violated at exactly one step\n"
      "                  \tof each unrolling (default checks at any step).\n"
      "  -h              \tPrints the command usage.\n\n");
  return 1;
}

