
#ifndef __ISC_INT_H__
#define __ISC_INT_H__

#include "compile/compile.h"
#include "opt/opt.h"
#include "trace/TraceManager.h"
#include "utils/defs.h"
#include "utils/NodeList.h"

/*---------------------------------------------------------------------------*/
/* Compilation Flags                                                         */
/*---------------------------------------------------------------------------*/

/* enables all the debugging and development helpers */
/* including automatic sanity checks */
#define ISC_DEBUG 0

/* enables checks for validation of the approximate environment */
/* transition relation: exact_trans -> approx_trans */
#define ISC_CONSISTENCY_CHECKS 0

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/
#define MAX_STR         65536

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
EXTERN FILE* nusmv_stdout;
EXTERN FILE* nusmv_stderr;
EXTERN cmp_struct_ptr cmps;

EXTERN TraceManager_ptr global_trace_manager;
EXTERN FlatHierarchy_ptr mainFlatHierarchy;
EXTERN FsmBuilder_ptr global_fsm_builder;
EXTERN DdManager* dd_manager;

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

/* commands (shell wrappers) */
int CommandIscCheckInvar ARGS((int argc, char **argv));

/* utils */
node_ptr isc_get_root_module_name ARGS((node_ptr n));
node_ptr isc_get_helper_var_name ARGS((int cnfLiteral));
int isc_get_original_be_index ARGS((node_ptr vname));
void isc_create_vars_and_layer ARGS((NodeList_ptr varlist));
bdd_ptr isc_get_vars_bdd ARGS((NodeList_ptr varlist));
be_ptr isc_get_partial_trans_be ARGS((BeEnc_ptr be_enc, Be_Manager_ptr be_mgr,
    node_ptr partial_trans_expr, int current_time));
boolean isc_is_int ARGS((char* str));
boolean is_helper_var_name_valid ARGS((node_ptr vname));
NodeList_ptr isc_get_new_helper_vars ARGS((NodeList_ptr sharedCnfVars,
    hash_ptr helpervars_map, hash_ptr varlist_map, BeEnc_ptr be_enc));

/* utils (debugging) see iscUtils.c */
#if 0
void isc_quantification_test ARGS((void));
#endif
void print_node_type ARGS((node_ptr n));
void print_type_structure ARGS((node_ptr node, int indent));

#endif /* __ISC_INT_H__ */

