
#include "iscInt.h"

#include "parser/symbols.h"
#include "utils/ustring.h"

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
node_ptr isc_get_helper_var_name(int beLiteral)
{
  nusmv_assert(beLiteral != 0);
  beLiteral = beLiteral < 0 ? -beLiteral : beLiteral;

  /* to avoid buffer overflow in sprintf */
  nusmv_assert(beLiteral < 1000000);

  char strbuffer[12];
  int written_chars = sprintf(strbuffer, "isc_%d", beLiteral);
  nusmv_assert(written_chars < 12);

  node_ptr result = find_node(DOT, Nil, find_node(ATOM,
      (node_ptr)find_string(strbuffer), Nil));

  nusmv_assert(isc_get_original_be_index(result) == beLiteral);

  return result;
}

/**Function********************************************************************

   Synopsis           [Retrieves the first part of a flattened variable name.]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
node_ptr isc_get_root_module_name (node_ptr n)
{
  nusmv_assert(Nil != n);

  switch (node_get_type(n)) {
  case ATOM:
    {
      return find_node(DOT, Nil, n);
    }
  case DOT:
    {
      if (Nil == car(n)) return n;
      else return isc_get_root_module_name(car(n));
    }
  case BIT:
    {
      nusmv_assert(cdr(n) == Nil);
      return isc_get_root_module_name(car(n));
    }
  default:
    {
      error_unreachable_code();
      break;
    }
  }

  error_unreachable_code();
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
int isc_get_original_be_index(node_ptr vname)
{
  nusmv_assert(is_helper_var_name_valid(vname));

  char* txt = get_text((string_ptr)cadr(vname));
  txt = txt + sizeof(char) * 4;
  return atoi(txt);
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
void isc_create_vars_and_layer(NodeList_ptr varlist)
{
  if (NodeList_get_length(varlist) == 0) return;

  SymbTable_ptr symb_table = Compile_get_global_symb_table();

  SymbLayer_ptr isc_var_layer =
      SymbTable_create_layer(symb_table, (char*)NULL, SYMB_LAYER_POS_BOTTOM);

  ListIter_ptr iter;
  NODE_LIST_FOREACH(varlist, iter) {
    node_ptr cur_name = NodeList_get_elem_at(varlist, iter);
    nusmv_assert(is_helper_var_name_valid(cur_name));
    SymbLayer_declare_state_var(isc_var_layer, cur_name,
        SymbType_create(SYMB_TYPE_BOOLEAN, Nil));
  }

  /* layer is not commited to be_enc since these helper variables are not used
   * in a be context. */

  /* bdd_enc is a subclass of bool_enc */
  BoolEnc_ptr bool_enc = Enc_get_bool_encoding();
  BaseEnc_commit_layer(BASE_ENC(bool_enc), SymbLayer_get_name(isc_var_layer));

  BddEnc_ptr bdd_enc = Enc_get_bdd_encoding();
  BaseEnc_commit_layer(BASE_ENC(bdd_enc), SymbLayer_get_name(isc_var_layer));
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
NodeList_ptr isc_get_new_helper_vars(NodeList_ptr sharedCnfVars,
    hash_ptr helpervars_map, hash_ptr varlist_map, BeEnc_ptr be_enc)
{
  Be_Manager_ptr be_mgr = BeEnc_get_be_manager(be_enc);

  NodeList_ptr new_vars = NodeList_create();
  nusmv_assert(NodeList_get_length(sharedCnfVars) > 0);

  ListIter_ptr iter;
  NODE_LIST_FOREACH(sharedCnfVars, iter) {

    int cnfLit = NODE_TO_INT(NodeList_get_elem_at(sharedCnfVars, iter));
    nusmv_assert(cnfLit > 0);

    int beLit = Be_CnfLiteral2BeLiteral(be_mgr, cnfLit);
    nusmv_assert(beLit > 0);

    int beIdx = Be_BeLiteral2BeIndex(be_mgr, beLit);
    nusmv_assert(beIdx > 0);

    node_ptr vkey = NODE_FROM_INT(beIdx);
    node_ptr vname_in_map = find_assoc(helpervars_map, vkey);

    if (Nil == vname_in_map) {
      node_ptr vname = isc_get_helper_var_name(beIdx);

      /* update hash and var list*/

      insert_assoc(helpervars_map, vkey, vname);
      NodeList_append(new_vars, vname);

      /* update varlist map */

      int time = BeEnc_index_to_time(be_enc, beIdx);

      NodeList_ptr varlist =
          NODE_LIST(find_assoc(varlist_map, NODE_FROM_INT(time)));
      
      if (NODE_LIST(NULL) == varlist) {
        varlist = NodeList_create();
        insert_assoc(varlist_map, NODE_FROM_INT(time), (node_ptr)varlist);
      }

      NodeList_append(varlist, vname);
    }
  }

  return new_vars;
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
bdd_ptr isc_get_vars_bdd(NodeList_ptr varlist)
{
  BddEnc_ptr bdd_enc = Enc_get_bdd_encoding();
  nusmv_assert(NODE_LIST(NULL) != varlist);
  nusmv_assert(NodeList_get_length(varlist) > 0);

  bdd_ptr vars_bdd = bdd_true(dd_manager);

  ListIter_ptr iter;
  NODE_LIST_FOREACH(varlist, iter) {
    node_ptr vname = NodeList_get_elem_at(varlist, iter);
    bdd_ptr cur_var = BddEnc_expr_to_bdd(bdd_enc, EXPR(vname), Nil);
    bdd_and_accumulate(dd_manager, &vars_bdd, cur_var);
    bdd_free(dd_manager, cur_var);
  }

  #if ISC_DEBUG
  {
    if (opt_verbose_level_gt(OptsHandler_get_instance(), 2)) {
      fprintf(nusmv_stderr, "vars as bdd: ");
      Expr_ptr exp = BddEnc_bdd_to_expr(bdd_enc, vars_bdd);
      print_node(nusmv_stderr, exp);
      fprintf(nusmv_stderr, "\n");
    }
  }
  #endif

  return vars_bdd;
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
be_ptr isc_get_partial_trans_be(BeEnc_ptr be_enc, Be_Manager_ptr be_mgr,
    node_ptr T, int current_time)
{
  be_ptr res;
  be_ptr res_if;
  be_ptr res_then;
  be_ptr res_else;
  be_ptr res_left;
  be_ptr res_right;

  nusmv_assert(Nil != T);

  if ((find_node(TRUEEXP, Nil, Nil)  == T)) return Be_Truth(be_mgr);
  if ((find_node(FALSEEXP, Nil, Nil) == T)) return Be_Falsity(be_mgr);

  switch (node_get_type(T)) {
  case DOT:
    {
      int beIdx = isc_get_original_be_index(T);
      nusmv_assert(beIdx > 0);
      nusmv_assert(!BeEnc_is_index_untimed(be_enc, beIdx));

      int beIdx_time = BeEnc_index_to_time(be_enc, beIdx);
      /* variables with other timings should have been removed by existential
       * quantification */
      nusmv_assert((beIdx_time == current_time) || (beIdx_time == (current_time + 1)));
      /* input variables should have been removed for t+1 */
      if (BeEnc_is_index_input_var(be_enc, beIdx)) {
        nusmv_assert(beIdx_time == current_time);
      }

      int beIdx_untimed = BeEnc_index_to_untimed_index(be_enc, beIdx);
      nusmv_assert(BeEnc_is_index_untimed(be_enc, beIdx_untimed));
      nusmv_assert(BeEnc_is_index_untimed_curr(be_enc, beIdx_untimed) ||
                   BeEnc_is_index_untimed_input(be_enc, beIdx_untimed));

      res = BeEnc_index_to_var(be_enc, beIdx_untimed);

      /* state variables at t+1 become next-state variables */
      if (beIdx_time != current_time) {
        nusmv_assert(BeEnc_is_index_untimed_curr(be_enc, beIdx_untimed));
        res = BeEnc_var_curr_to_next(be_enc, res);
      }

      break;
    }
  case NOT:
    {
      nusmv_assert(Nil == cdr(T));
      res_left = isc_get_partial_trans_be(be_enc, be_mgr, car(T), current_time);
      res = Be_Not(be_mgr, res_left);
      break;
    }
  case AND:
    {
      res_left  = isc_get_partial_trans_be(be_enc, be_mgr, car(T), current_time);
      res_right = isc_get_partial_trans_be(be_enc, be_mgr, cdr(T), current_time);
      res = Be_And(be_mgr, res_left, res_right);
      break;
    }
  case OR:
    {
      res_left  = isc_get_partial_trans_be(be_enc, be_mgr, car(T), current_time);
      res_right = isc_get_partial_trans_be(be_enc, be_mgr, cdr(T), current_time);
      res = Be_Or(be_mgr, res_left, res_right);
      break;
    }
  case CASE:
    {
      nusmv_assert(node_get_type(car(T)) == COLON);
      res_if   = isc_get_partial_trans_be(be_enc, be_mgr, caar(T), current_time);
      res_then = isc_get_partial_trans_be(be_enc, be_mgr, cdar(T), current_time);
      res_else = isc_get_partial_trans_be(be_enc, be_mgr, cdr(T),  current_time);
      res = Be_Ite(be_mgr, res_if, res_then, res_else);
      break;
    }
  default:
    error_unreachable_code();
  }

  return res;
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
boolean isc_is_int(char* str)
{
  while ((char*)NULL != str && '\0' != str[0]) {
    if (!isdigit((int)str[0])) return false;
    str += sizeof(char);
  }
  return true;
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
boolean is_helper_var_name_valid(node_ptr vname)
{
  if (Nil == vname) return false;
  if (DOT != node_get_type(vname)) return false;
  if (Nil != car(vname)) return false;
  if (ATOM != node_get_type(cdr(vname))) return false;
  if (Nil != cddr(vname)) return false;

  char* txt = get_text((string_ptr)cadr(vname));

  if ((char*)NULL == txt) return false;
  if (strlen(txt) <= 4) return false;
  if (0 != strncmp("isc_", txt, 4)) return false;

  txt = txt + sizeof(char) * 4;
  if (!isc_is_int(txt)) return false;

  int result = atoi(txt);
  if (result <= 0) return false;

  return true;
}


#if 0

/**Function********************************************************************

   Synopsis           [debugging/development code]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/

void isc_quantification_test()
{
  BddEnc_ptr bdd_enc = Enc_get_bdd_encoding();

  node_ptr v10_name = isc_get_helper_var_name(10);
  node_ptr v20_name = isc_get_helper_var_name(20);
  node_ptr v30_name = isc_get_helper_var_name(30);
  node_ptr v40_name = isc_get_helper_var_name(40);
  
  NodeList_ptr varlist = NodeList_create();
  NodeList_append(varlist, v10_name);
  NodeList_append(varlist, v20_name);
  NodeList_append(varlist, v30_name);
  NodeList_append(varlist, v40_name);

  isc_create_vars_and_layer(varlist);

  NodeList_destroy(varlist);
  
  Expr_ptr ev10 = EXPR(v10_name);
  Expr_ptr ev20 = EXPR(v20_name);
  Expr_ptr ev30 = EXPR(v30_name);
  Expr_ptr ev40 = EXPR(v40_name);

  Expr_ptr expr_before = 
    Expr_or(
      Expr_and(ev10, ev20),
      Expr_or(
        Expr_and(
          Expr_not(ev10),
          ev30),
        Expr_and(
          ev30,
          Expr_and(
            ev40,
            Expr_not(ev20)))));

  printf("\nexpr before:\n");
  print_node(nusmv_stdout, (node_ptr)expr_before);
  printf("\n");

  bdd_ptr expr_as_bdd = BddEnc_expr_to_bdd(bdd_enc, expr_before, Nil);

  Expr_ptr expr_after = BddEnc_bdd_to_expr(bdd_enc, expr_as_bdd);

  printf("\nexpr after:\n");
  print_node(nusmv_stdout, (node_ptr)expr_after);
  printf("\n");

  varlist = NodeList_create();
  NodeList_append(varlist, v40_name);

  bdd_ptr vars_to_be_removed = isc_get_vars_bdd(varlist);

  NodeList_destroy(varlist);

  bdd_ptr quantified_bdd = bdd_forsome(dd_manager, expr_as_bdd, vars_to_be_removed);

  Expr_ptr expr_quantified = BddEnc_bdd_to_expr(bdd_enc, quantified_bdd);

  printf("\nquantified bdd-expr:\n");
  print_node(nusmv_stdout, (node_ptr)expr_quantified);
  printf("\n");
}

#endif

static void print_indent(int indent)
{
  if (indent < 2) return;
  while (indent-- > 0) printf(".");
}

void print_node_type(node_ptr n)
{
  switch (node_get_type(n)) {
  case ATOM:
    printf("ATOM");
    break;
  case DOT:
    printf("DOT");
    break;
  case CASE:
    printf("CASE");
    break;
  case COLON:
    printf("COLON");
    break;
  case BIT:
    printf("BIT");
    break;
  case AND:
    printf("AND");
    break;
  case OR:
    printf("OR");
    break;
  case NOT:
    printf("NOT");
    break;
  case FALSEEXP:
    printf("FALSE");
    break;
  case TRUEEXP:
    printf("TRUE");
    break;
  case CONS:
    printf("CONS");
    break;
  default:
    printf("%d", node_get_type(n));
    break;
  }
}

void print_type_structure(node_ptr node, int indent)
{
  print_indent(indent);

  if (Nil == node) {
    printf("Nil\n");
    return;
  }

  if (ATOM == node_get_type(node)) {
    printf("ATOM (%s)\n",
        get_text((string_ptr)car(node)));
    return;
  }

  if (BIT == node_get_type(node)) {
    printf("BIT (%d)", NODE_TO_INT(cdr(node)));
  }
  else {
    print_node_type(node);
  }

  printf("\n");

  /* print left subtree */
  print_type_structure(car(node), indent+2);

  /* print right subtree */
  print_type_structure(cdr(node), indent+2);
}

