/**************************************************************************************************

Solver_C.h

C-wrapper for Solver.h

**************************************************************************************************/

#ifndef SOLVER_C_h
#define SOLVER_C_h


//=================================================================================================
// Solver -- the main class:

#define MiniSat_ptr void *

MiniSat_ptr MiniSat_Create();
void MiniSat_Delete(MiniSat_ptr);
int MiniSat_Nof_Variables(MiniSat_ptr);
int MiniSat_Nof_Clauses(MiniSat_ptr);
int MiniSat_New_Variable(MiniSat_ptr);
int MiniSat_Add_Clause(MiniSat_ptr, int *clause_lits, int num_lits);
int MiniSat_Solve(MiniSat_ptr);
int MiniSat_Solve_Assume(MiniSat_ptr, int nof_assumed_lits, int *assumed_lits);
int MiniSat_simplifyDB(MiniSat_ptr);
int MiniSat_Get_Value(MiniSat_ptr, int var_num);

/* new functions for itp support */

void MiniSat_EnableInterpolation(MiniSat_ptr);
int MiniSat_InterpolationIsEnabled(MiniSat_ptr);
void MiniSat_EnableVerboseProofLogging(MiniSat_ptr);
void MiniSat_DisableVerboseProofLogging(MiniSat_ptr);

void MiniSat_ComputeVariableLists(MiniSat_ptr);
void MiniSat_PrintVariableLists(MiniSat_ptr);

void MiniSat_MarkProblemSplit(MiniSat_ptr, int i);

/* interpolation data structures and ifc functions */

typedef enum ItpNodeType_TAG { PIVOT_A = 0, PIVOT_B = 1, PIVOT_AB = 2, ROOT_A = 3, ROOT_B = 4 } ItpNodeType;

typedef int Var;

struct ItpNode {

  ItpNodeType nType;
  Var pivotVariable;

  struct ItpNode * parentNegative;  // parent where pivot variable occurs negatively
  struct ItpNode * parentPositive;  // parent where pivot variable occurs positively
  
  void * interpolant;               // the resulting interpolant for this node (= be_ptr)
                                    // to be used on nusmv side
	int k;							// the k^th step to which this clause belongs to
	int *clause;                    // if self is a root node, then this is the clause
	int clause_size;                // length of the clause
};

struct ItpNode * MiniSat_GetInterpolantGraph (MiniSat_ptr);

//=================================================================================================
#endif

